/******************************************************************************
 * music.c
 * SASS Compiler for CoreTone, music parsing and track generation.
 *-----------------------------------------------------------------------------
 * Copyright (C) 2015 - 2016 Osman Celimli
 * For conditions of distribution and use, see copyright notice in ctcc.c
 ******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdint.h>

#include "../types.h"
#include "../CoreTone/sample.h"
#include "../CoreTone/channel.h"
#include "../CoreTone/music.h"

#include "useful.h"
#include "sample.h"
#include "macro.h"
#include "music.h"

/******************************************************************************
 * Text Tables
 ******************************************************************************/
extern const char szERR_NoMainMem[];
extern const char szERR_Range[];
extern const char szERR_Syntax[];
extern const char szERR_EarlyEnd[];
extern const char szERR_NoOpen[];
extern const char szERR_NoDest[];

extern const char szERR_Garbage[];

const char szMUS_Parse[] =		"Parsing music script file : ";
const char szMUS_Make[] =		"Linking and saving music track to : ";

const char szMUS_NoMain[] =		"WARNING- No main block found for channel : ";
const char szMUS_ResCount[] =	"WARNING- Loop count of -128 reserved for CALLs\n";

const char szMUS_TwoLabels[] =	"ERROR- Block already defined with label : ";
const char szMUS_BadLabel[] =	"ERROR- Invalid music block label in : ";
const char szMUS_NoBrace[] =	"ERROR- Expecting '{' at block start : ";
const char szMUS_BadPri[] =		"ERROR- Music block priority out of range in : ";
const char szMUS_TwoMains[] =	"ERROR- Second main block declaration in : ";
const char szMUS_BlockBusy[] =	"ERROR- Unexpected end of music block in : ";
const char szMUS_LostBlock[] =	"ERROR- Unable to resolve reference to block : ";

const char szMUS_BadInst[] =	"ERROR- Instrument not found : ";
const char szMUS_BadNote[] =	"ERROR- Unable to parse note : ";
const char szMUS_NoteFreq[] =	"ERROR- Note frequency out of range (c.0 to g.10)\n";

const char szMUS_BadWrite[] =	"ERROR- Couldn't write to output file : ";

const char szMUS_WhiteSpace[] =	" ,\t\n";

/******************************************************************************
 * Globals and Externs
 ******************************************************************************/
/* Parsing State
 *----------------------------------------------------------------------------*/
char *pszMusFile = NULL;
char *pszMusBuf = NULL;
uint32_t uiMusLine = 0;
uint32_t uiMusOffset = 0;

uint32_t uiCurChannel = 0;


/* Active Blocks and Tags
 *----------------------------------------------------------------------------*/
MusicBlock_t *paMusBlocks = NULL;
uint32_t uiBlocks = 0,uiBlockSlots = 0;
MusicTag_t *paMusTags = NULL;
uint32_t uiTags = 0,uiTagSlots = 0;
MusicTrack_t *paMusTracks = NULL;
uint32_t uiTracks = 0, uiTrackSlots;

uint32_t uiCurBlock = 0;
uint32_t uiBlockBase = 0;


/* Active Instrument and Note
 *----------------------------------------------------------------------------*/
uint32_t uiCurInst = 0;
int32_t iInstSet = 0;

int32_t iNoteOn = 0;
int32_t iNoiseInst = 0;


/******************************************************************************
 * !!!!----                  MUSIC SCRIPT COMMANDS                   ----!!!!
 ******************************************************************************/
typedef int32_t (*MusCom_t)(MusicBlock_t *pBlock);

/* using,u instrument
 -----------------------------------------------------------------------------*/
int32_t musCom_using(MusicBlock_t *pBlock)
{
	MacroBlock_t *pMacro;
	uint32_t uiMacroNum;
	uint8_t ucMacroNum;

	char *pszSubTok;
	int32_t iStatus;


	if(NULL == (pszSubTok = strtok(NULL, szMUS_WhiteSpace)))
		return 1;

	if(mac_get(pszSubTok, &uiMacroNum, &pMacro))
	{
		fprintf(stderr, "%s%s\n", szMUS_BadInst, pszSubTok);
		return -1;
	}

	uiCurInst = uiMacroNum;
	ucMacroNum = uiMacroNum & 0xFF;
	iInstSet = -1;

	iStatus = mus_addByte(pBlock, CORETONE_MUSIC_SET_INSTRUMENT);
	iStatus |= mus_addByte(pBlock, ucMacroNum);
	return iStatus;
}

/* priority,pr value
 -----------------------------------------------------------------------------*/
int32_t musCom_priority(MusicBlock_t *pBlock)
{
	char *pszSubTok;

	int32_t iStatus,iVal;
	int8_t cPri;


	if(NULL == (pszSubTok = strtok(NULL, szMUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;

	if((iVal < MUSIC_PARSER_PRIORITY_MIN) || (iVal > MUSIC_PARSER_PRIORITY_MAX))
	{
		fprintf(stderr, "%s", szERR_Range);
		return 1;
	}

	cPri = iVal & 0xFF;
	iStatus = mus_addByte(pBlock, CORETONE_MUSIC_SET_PRIORITY);
	iStatus |= mus_addByte(pBlock, cPri);
	return iStatus;
}

/* pan,p value
 -----------------------------------------------------------------------------*/
int32_t musCom_pan(MusicBlock_t *pBlock)
{
	char *pszLeftTok,*pszRightTok;

	int32_t iStatus,iLeft,iRight;
	uint8_t ucLeft,ucRight;


	pszLeftTok = strtok(NULL, szMUS_WhiteSpace);
	pszRightTok = strtok(NULL, szMUS_WhiteSpace);

	if((NULL == pszLeftTok) || (NULL == pszRightTok))
		return 1;
	if(strtoint(pszLeftTok, &iLeft) || strtoint(pszRightTok, &iRight))
		return 1;

	iStatus = (iLeft < MUSIC_PARSER_PANNING_MIN) || (iLeft > MUSIC_PARSER_PANNING_MAX)
		|| (iRight < MUSIC_PARSER_PANNING_MIN) || (iRight > MUSIC_PARSER_PANNING_MAX);
	if(iStatus)
	{
		fprintf(stderr, "%s", szERR_Range);
		return 1;
	}

	ucLeft = iLeft & 0xFF;
	ucRight = iRight & 0xFF;
	iStatus = mus_addByte(pBlock, CORETONE_MUSIC_SET_PANNING);
	iStatus |= mus_addByte(pBlock, ucLeft);
	iStatus |= mus_addByte(pBlock, ucRight);
	return iStatus;
}

/* pitch,pt offset [adjustment]
 -----------------------------------------------------------------------------*/
int32_t musCom_pitch(MusicBlock_t *pBlock)
{
	char *pszOffTok,*pszAdjTok;
	int16p16_t stOff,stAdj;
	int8p8_t stWrite;

	int32_t iStatus;


	pszOffTok = strtok(NULL, szMUS_WhiteSpace);
	pszAdjTok = strtok(NULL, szMUS_WhiteSpace);
	if(NULL == pszOffTok)
		return 1;
	if(parse16p16(&stOff, pszOffTok))
		return 1;

	if(NULL != pszAdjTok)
	{
		if(parse16p16(&stAdj, pszAdjTok))
			return 1;
	}
	else
	{
		stAdj.iWhole = 0;
	}

	iStatus = mus_addByte(pBlock, CORETONE_MUSIC_PITCH);

	stWrite.usWhole = stOff.usPair.usLo;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);
	stWrite.usWhole = stOff.usPair.usHi;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);

	stWrite.usWhole = stAdj.usPair.usLo;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);
	stWrite.usWhole = stAdj.usPair.usHi;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);
	return iStatus;
}

/* mood,m value
 -----------------------------------------------------------------------------*/
int32_t musCom_mood(MusicBlock_t *pBlock)
{
	char *pszSubTok;

	int16p16_t stVal;
	int8p8_t stWrite;

	int32_t iStatus;


	if(NULL == (pszSubTok = strtok(NULL, szMUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &(stVal.iWhole)))
		return 1;

	iStatus = mus_addByte(pBlock, CORETONE_MUSIC_SET_MOOD);

	stWrite.usWhole = stVal.usPair.usLo;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);
	stWrite.usWhole = stVal.usPair.usHi;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);
	return iStatus;
}


/* loop,l count
 -----------------------------------------------------------------------------*/
int32_t musCom_loop(MusicBlock_t *pBlock)
{
	char *pszSubTok;

	uint8_t ucCount;
	int32_t iStatus,iVal;


	if(NULL == (pszSubTok = strtok(NULL, szMUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < MUSIC_PARSER_LOOP_MIN) || (iVal > MUSIC_PARSER_LOOP_MAX))
	{
		fprintf(stderr, "%s", szERR_Range);
		return 1;
	}

	if(CORETONE_MUSIC_CALL_TAG == iVal)
	{
		fprintf(stderr, "%s", szMUS_ResCount);
	}

	ucCount = iVal & 0xFF;
	iStatus = mus_addByte(pBlock, CORETONE_MUSIC_LOOP_START);
	iStatus |= mus_addByte(pBlock, ucCount);
	return iStatus;
}

/* endloop,el
 -----------------------------------------------------------------------------*/
int32_t musCom_endloop(MusicBlock_t *pBlock)
{
	return mus_addByte(pBlock, CORETONE_MUSIC_LOOP_END);
}

/* call,c label
 -----------------------------------------------------------------------------*/
int32_t musCom_call(MusicBlock_t *pBlock)
{
	char *pszSubTok;
	int32_t iStatus;

	if(mus_addByte(pBlock, CORETONE_MUSIC_CALL))
		return -1;
	if(NULL == (pszSubTok = strtok(NULL, szMUS_WhiteSpace)))
		return 1;
	if(mus_addCallTag(pszSubTok, uiMusOffset, pszMusFile, uiMusLine))
		return -1;

	iStatus = mus_addByte(pBlock, 0);
	iStatus |= mus_addByte(pBlock, 0);
	iStatus |= mus_addByte(pBlock, 0);
	iStatus |= mus_addByte(pBlock, 0);
	return iStatus;
}

/* return,r
 -----------------------------------------------------------------------------*/
int32_t musCom_return(MusicBlock_t *pBlock)
{
	return mus_addByte(pBlock, CORETONE_MUSIC_RETURN);
}

/* break,b
 -----------------------------------------------------------------------------*/
int32_t musCom_break(MusicBlock_t *pBlock)
{
	return mus_addByte(pBlock, CORETONE_MUSIC_BREAK);
}


/* wait,w ticks
 -----------------------------------------------------------------------------*/
int32_t musCom_wait(MusicBlock_t *pBlock)
{
	char *pszSubTok;

	uint8_t aucDelay[4];
	int32_t iVal,iDiv;

	int32_t iStatus;
	uint32_t uiX,uiY;


	if(NULL == (pszSubTok = strtok(NULL, szMUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if(iVal < MACRO_PARSER_WAIT_MIN)
	{
		fprintf(stderr, "%s", szERR_Range);
		return 1;
	}

	/**
	 * Delay commands are a special case and operate similarly to MIDI's
	 * varlength delays. Any command byte with its MSB set is interpreted
	 * as the start of a delay string which continues until a command
	 * with its MSB clear is encountered.
	 *
	 * The seven remaining bits of each byte in the delay string are used
	 * as the delay count itself, and are shifted into bits 7-0, 14-8,
	 * 21-15, or 28-22 of the accumulated delay count. If the delay is
	 * larger than four bytes' worth of ticks, it can be followed by
	 * a consecutive delay string.
	 *
	 * If the last command we processed was another delay, we need to
	 * backtrack the script and merge its delay count with our new one.
	 */
	if(pBlock->iWasWaiting)
	{
		iVal += pBlock->uiDelayAcc;
		pBlock->uiScriptLen -= pBlock->uiDelayLen;
		uiMusOffset -= pBlock->uiDelayLen;

		pBlock->uiDelayAcc = 0;
		pBlock->uiDelayLen = 0;
		pBlock->iWasWaiting = 0;
	}

	iStatus = 0;
	while(iVal > 0)
	{
		memset(&aucDelay, 0, (sizeof(uint8_t) * 4));
		pBlock->uiDelayAcc = iVal;
		uiY = 1;
		if(iVal >= 0x200000)
		{
			/*---- Bits 28-22 ----*/
			iDiv = (iVal / 0x200000) & 0x7F;
			iVal -= (iDiv * 0x200000);

			aucDelay[3] = iDiv;
			uiY = 4;
		}
		if(iVal >= 0x4000)
		{
			/*---- Bits 21-15 ----*/
			iDiv = (iVal / 0x4000) & 0x7F;
			iVal -= (iDiv * 0x4000);

			aucDelay[2] = iDiv;
			if(uiY < 3) uiY = 3;
		}
		if(iVal >= 0x80)
		{
			/*---- Bits 14-8 ----*/
			iDiv = (iVal / 0x80) & 0x7F;
			iVal -= (iDiv * 0x80);

			aucDelay[1] = iDiv;
			if(uiY < 2) uiY = 2;
		}
		if(iVal > 0x00)
		{
			/*---- Bits 7-0 ----*/
			iDiv = iVal & 0x7F;
			iVal -= iDiv;

			aucDelay[0] = iDiv;
		}

		pBlock->iWasWaiting = -1;
		pBlock->uiDelayLen = uiY;
		for(uiX = 0; uiX < uiY; uiX++)
		{
			iStatus |= mus_addByte(pBlock,
				(uint8_t)(aucDelay[uiX] | CORETONE_MUSIC_WAIT));
		}
	}

	return iStatus;
}

/* rest,r ticks
 -----------------------------------------------------------------------------*/
int32_t musCom_rest(MusicBlock_t *pBlock)
{
	int32_t iStatus;

	iStatus = 0;
	if(iNoteOn)
	{
		iNoteOn = 0;
		iStatus |= mus_addByte(pBlock, CORETONE_MUSIC_NOTE_OFF);
	}

	iStatus |= musCom_wait(pBlock);
	return iStatus;
}

/* end,e
 -----------------------------------------------------------------------------*/
int32_t musCom_end(MusicBlock_t *pBlock)
{
	int iStatus;

	iStatus = 0;
	if(iNoteOn)
	{
		iNoteOn = 0;
		iStatus |= mus_addByte(pBlock, CORETONE_MUSIC_NOTE_OFF);
	}

	iStatus |= mus_addByte(pBlock, MUSIC_PARSER_PRIORITY_OFF);
	iStatus |= mus_addByte(pBlock, 0);
	return iStatus;
}


/******************************************************************************
 * !!!!----                     COMMAND TABLE                        ----!!!!
 ******************************************************************************/
char *apszMusCom_nameLong[] = {
	"using", "priority",
	"pan", "pitch", "mood",
	"loop", "endloop", "call", "return", "break",
	"wait", "rest", "end",
	NULL
};

char *apszMusCom_nameShort[] = {
	"u", "pr",
	"p", "pt", "m",
	"l", "el", "c", "rt", "b",
	"w", "r", "e",
	NULL
};

MusCom_t aMusCom_addr[] = {
	musCom_using, musCom_priority,
	musCom_pan, musCom_pitch, musCom_mood,
	musCom_loop, musCom_endloop, musCom_call, musCom_return, musCom_break,
	musCom_wait, musCom_rest, musCom_end,
	NULL
};


/******************************************************************************
 * !!!!----                  MUSIC SCRIPT PARSING                    ----!!!!
 ******************************************************************************/
const char *aszMidi_NoteTable[] = {
	"c.","cs","d.","ds","e.","f.",
	"fs","g.","gs","a.","as","b.",
	NULL
};

/* int32_t mus_addNote(MusicBlock_t *pBlock, char *pszNote)
 *  Attempt to parse the string pszNote as a note and add it to the current
 * music stream, returns nonzero if any errors occur.
 *----------------------------------------------------------------------------*/
int32_t mus_addNote(MusicBlock_t *pBlock, char *pszNote)
{
	char szBase[4],*pszOctave;
	uint32_t uiBase,uiOctave,uiNote;
	uint32_t uiLen;
	uint8_t ucNote;

	int32_t iStatus;


	/**
	 * Notes use a three characer format arranged as : nXo
	 *  n = Note (a,g,f,etc)
	 *  X = Flat (b) or Sharp (s)
	 *  o = Octave (0+)
	 *
	 * CoreTone note bytes are encoded identically to General MIDI's in A440,
	 * so we just need to extract the note and octave from the source string
	 * then combine them into the final value.
	 *
	 * Base note values are pulled from the array above, yielding a value of
	 * zero to eleven. Each octave is adds twelve to this base value, i.e.
	 * cs.3 would yield : 1 + 36 = 37.
	 */
	uiLen = strlen(pszNote);
	if(uiLen < 3)
	{
		fprintf(stderr, "%s%s\n", szMUS_BadNote, pszNote);
		return -1;
	}

	szBase[0] = pszNote[0];
	szBase[1] = pszNote[1];
	szBase[2] = '\0';
	for(uiBase = 0; uiBase < 12; uiBase++)
	{
		if(0 == strcmp(szBase, aszMidi_NoteTable[uiBase]))
		{
			break;
		}
	}

	if(12 == uiBase)
	{
		fprintf(stderr, "%s%s\n", szMUS_BadNote, pszNote);
		return -1;
	}

	pszOctave = pszNote + 2;
	if(1 != sscanf(pszOctave, "%u", &uiOctave))
	{
		fprintf(stderr, "%s%s\n", szMUS_BadNote, pszNote);
		return -1;
	}

	uiNote = uiBase + (uiOctave * 12);
	if(uiNote > 127)
	{
		fprintf(stderr, "%s\n", szMUS_NoteFreq);
		return -1;
	}

	ucNote = uiNote;
	iStatus = mus_addByte(pBlock, CORETONE_MUSIC_NOTE_ON);
	iStatus |= mus_addByte(pBlock, ucNote);
	iStatus |= musCom_wait(pBlock);
	iNoteOn = -1;
	return iStatus;
}

/* int32_t mus_parse(FILE *pSASS, char *pszFileName)
 *  Parse the SASS music script in the file pSASS whose name is stored in
 * pszFileName. The file must be opened beforehand. The current parsing channel
 * in uiCurChannel will be autoincremented after the entire file has
 * been traversed.
 *
 *  Will return a nonzero value if any errors are encountered during parsing.
 *----------------------------------------------------------------------------*/
int32_t mus_parse(FILE *pSASS, char *pszFileName)
{
	MusicBlock_t *pCurBlock = NULL;
	int32_t iNewPri = 0;

	int32_t iAborted = -1,iStatus = 0;
	int32_t iBlockBusy = 0,iBlockBraced = 0;
	char *pszMainTok = NULL,*pszSubTok = NULL;
	uint32_t uiX;

	/**
	 * Reinitialize select globals (since this will be called once for each
	 * SASS source) and default the temporary use block.
	 */
	pszMusFile = pszFileName;
	uiCurInst = 0;
	iInstSet = 0;
	iNoteOn = 0;

	fprintf(stdout, "%s%s\n", szMUS_Parse, pszFileName);
	if(0 == uiCurChannel)
	{
		paMusBlocks = malloc(sizeof(MusicBlock_t) * MUSIC_PARSER_BLOCK_LEN);
		paMusTags = malloc(sizeof(MusicTag_t) * MUSIC_PARSER_TAG_LEN);
		paMusTracks = malloc(sizeof(MusicTrack_t) * MUSIC_PARSER_TRACK_LEN);

		if((NULL == paMusBlocks) || (NULL == paMusTags) || (NULL == paMusTracks))
		{
			fprintf(stderr, "%s", szERR_NoMainMem);
			return -1;
		}
		uiBlockSlots = MUSIC_PARSER_BLOCK_LEN;
		uiTagSlots = MUSIC_PARSER_TAG_LEN;
		uiTrackSlots = MUSIC_PARSER_TRACK_LEN;
	}

	paMusTracks[uiTracks].iFoundMain = 0;
	paMusTracks[uiTracks].iMainPriority = MUSIC_PARSER_PRIORITY_OFF;
	paMusTracks[uiTracks].uiMainOffset = 0;
	pCurBlock = paMusBlocks + uiBlocks;
	pCurBlock->pacStream = NULL;
	pCurBlock->pszName = NULL;


	/**
	 * !!-- PARSE START --!!
	 * Allocate our linebuffer if that hasn't bee done already, then begin
	 * traversing the SASS source line by line.
	 */
	if(NULL == pszMusBuf)
	{
		pszMusBuf = malloc(sizeof(char) * MUSIC_PARSER_LINEBUF_LEN);
		if(NULL == pszMusBuf)
		{
			fprintf(stderr, "%s", szERR_NoMainMem);
			return -1;
		}
	}

	for(uiMusLine = 1; ; uiMusLine++)
	{
		/**
		 * Fetch our current line from the source file, stripping out any
		 * ;comments and tokenizing the remaining characters.
		 */
		if(fgets(pszMusBuf, MUSIC_PARSER_LINEBUF_LEN, pSASS) == NULL)
		{
			iAborted = 0;
			break;
		}
		if(strcut(pszMusBuf, ";", '\0')) break;
		pszMainTok = strtok(pszMusBuf, szMUS_WhiteSpace);

		if(iBlockBusy && (NULL != pszMainTok))
		{
			/**
			 * !!-- BLOCK PARSING --!!
			 * -----------------------
			 * Once a block has been opened with a label and curlybrace {, we
			 * can actually process its body.
			 */
			if(iBlockBraced)
			{
				if(0 == strcmp(pszMainTok, "}"))
				{
					/**
					 * Once we've hit our closing curlybrace }, it's time to
					 * add the block to our current data stream.
					 */
					iBlockBusy = 0;
					iBlockBraced = 0;
					if(tealloc(&paMusBlocks, &uiBlockSlots, (uiBlocks + 2), sizeof(MusicBlock_t)))
						break;

					uiBlocks++;
					pCurBlock = paMusBlocks + uiBlocks;
					pCurBlock->pacStream = NULL;
					pCurBlock->pszName = NULL;
				}
				else
				{
					/**
					 * Entries in the block body could be either notes or a
					 * command, we'll check for commands first and then see
					 * if the thing is a note if none of the commands in our
					 * table match up with the main token.
					 */
					for(uiX = 0; (NULL != aMusCom_addr[uiX]); uiX++)
					{
						if((0 == strcmp(apszMusCom_nameLong[uiX], pszMainTok)) ||
						   (0 == strcmp(apszMusCom_nameShort[uiX], pszMainTok)))
						{
							break;
						}
					}

					/**
					 * Both the command handlers and the note parser return a
					 * nonzero value in case if error.
					 *
					 * A positive nonzero value indicates a syntax failure
					 * error a negative value indicates an allocation / system
					 * failure.
					 */
					if(musCom_wait != aMusCom_addr[uiX])
					{
						pCurBlock->iWasWaiting = 0;
					}

					iStatus = (NULL != aMusCom_addr[uiX])
						? (*aMusCom_addr[uiX])(pCurBlock)
						: mus_addNote(pCurBlock, pszMainTok);
					if(iStatus > 0)
					{
						fprintf(stderr, "%s%s,%u\n", szERR_Syntax,
							pszFileName, uiMusLine);
						break;
					}
					else if(iStatus < 0)
					{
						break;
					}
				}
			}
			else
			{
				if(0 == strcmp(pszMainTok, "{"))
				{
					iBlockBraced = -1;
				}
				else
				{
					fprintf(stderr, "%s%s,%u\n", szMUS_NoBrace,
						pszFileName, uiMusLine);
					break;
				}
			}
		}
		else if(NULL != pszMainTok)
		{
			/**
			 * !!-- BLOCK DETECTION --!!
			 * -------------------------
			 * We're not currently parsing a block, so we need to try and find
			 * one with the format: label [priority]
			 */
			pszSubTok = strtok(NULL, szMUS_WhiteSpace);
			if(NULL != pszSubTok)
			{
				/**
				 * If we found two tokens, the second one will (or should) be
				 * the priority indicator for this file's main block.
				 */
				if(1 == sscanf(pszSubTok, "%d", &iNewPri))
				{
					if((iNewPri < MUSIC_PARSER_PRIORITY_MIN) || (iNewPri > MUSIC_PARSER_PRIORITY_MAX))
					{
						fprintf(stderr, "%s%s,%u\n", szMUS_BadPri,
							pszFileName, uiMusLine);
						break;
					}
					else
					{
						if(paMusTracks[uiTracks].iFoundMain)
						{
							fprintf(stderr, "%s%s,%u\n", szMUS_TwoMains,
								pszFileName, uiMusLine);
							break;
						}

						paMusTracks[uiTracks].iFoundMain = -1;
						paMusTracks[uiTracks].iMainPriority = iNewPri;
						paMusTracks[uiTracks].uiMainOffset = uiMusOffset;
					}
				}
				else
				{
					fprintf(stderr, "%s%s,%u\n", szMUS_BadLabel,
						pszFileName, uiMusLine);
					break;
				}
			}

			/**
			 * Check to see if a block with this one's label already exists.
			 * If not, we can set up the temporary use block and switch into
			 * block parsing mode for our next line.
			 */
			if(0 == mus_getBlock(pszMainTok, NULL, NULL))
			{
				fprintf(stderr, "%s%s in : %s,%u\n", szMUS_TwoLabels,
					pszMainTok, pszFileName, uiMusLine);

				break;
			}

			mus_prepBlock(pCurBlock, pszMainTok);
			iBlockBusy = -1;
			iBlockBraced = 0;
			iInstSet = 0;
			iNoteOn = 0;
		}

		/**
		 * Regardless of mode, check for garbage at the end of the current
		 * parsing line before advancing.
		 */
		pszSubTok = strtok(NULL, szMUS_WhiteSpace);
		if(NULL != pszSubTok)
		{
			fprintf(stderr, "%s%s,%u\n", szERR_Garbage,
				pszFileName, uiMusLine);
			break;
		}
	}

	/**
	 * Check to see if any errors were encountered during parsing, prepping
	 * a slot for another track in case we still have more SASS to compile.
	 * Warn the user if they never declared a main track for this file.
	 */
	iStatus = 0;
	if(iBlockBusy)
	{
		fprintf(stderr, "%s%s,%u\n", szMUS_BlockBusy,
			pszFileName, uiMusLine);
	}
	else if(!(iAborted | iBlockBusy))
	{
		if(!(paMusTracks[uiTracks].iFoundMain))
		{
			fprintf(stderr, "%s%s\n", szMUS_NoMain, pszFileName);
		}

		iStatus = tealloc(&paMusTracks, &uiTrackSlots,
			(uiTracks + 2), sizeof(MusicTrack_t));

		if(!iStatus)
		{
			uiTracks++;
		}
	}

	uiCurChannel++;
	return (iAborted | iBlockBusy | iStatus);
}

/* int32_t mus_make(FILE *pDest, char *pszFileName)
 *  Attempt to build and write a music binary for the currently populated track
 * and block tables. All CALLs and offsets will be resolved at this time.
 * The destination file pDest should be open for binary writing at the time
 * this function is called.
 *  Returns a nonzero value if anything went wrong during the build.
 *----------------------------------------------------------------------------*/
int32_t mus_make(FILE *pDest, char *pszFileName)
{
	MusicBlock_t *pTarget;
	uint8_t *pMusic = NULL;
	uint8_t *pDir = NULL, *pCall = NULL;

	uint32_t uiBlockBase,uiOff;
	uint32_t uiLen;

	uint32_t uiX,uiY,uiZ;
	int32_t iStatus = -1;

	/**
	 * !!!!---- PREBUILD ACTIVITIES ----!!!!
	 * Scan through the compiled blocks to calculate the amount of data
	 * required for the final music binary, then allocate it.
	 */
	fprintf(stdout, "%s%s\n", szMUS_Make, pszFileName);
	uiLen = CORETONE_MUSPAK_HEAD_SIZE + (uiTracks * CORETONE_MUSPAK_ENTRY_LEN);
	uiBlockBase = uiLen;
	uiOff = uiLen;
	for(uiX = 0; uiX < uiBlocks; uiX++)
	{
		uiLen += paMusBlocks[uiX].uiScriptLen;
	}

	pMusic = malloc(sizeof(uint8_t) * uiLen);
	if(NULL == pMusic)
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}
	pDir = pMusic + CORETONE_MUSPAK_HEAD_SIZE;


	/**
	 * !!!!---- SCRIPT COPY ----!!!!
	 * Copy the music scripts from their respective blocks to the target
	 * binary, preserving their order and offsets.
	 */
	strcpy(pMusic, CORETONE_MUSPAK_HEAD_MAGICWORD);
	memcpy(pMusic + CORETONE_MUSPAK_HEAD_TRACKS, &uiTracks, sizeof(uint32_t));
	for(uiX = 0; uiX < uiBlocks; uiX++)
	{
		memcpy(pMusic + uiOff, paMusBlocks[uiX].pacStream, 
			(paMusBlocks[uiX].uiScriptLen * sizeof(uint8_t)));

		uiOff += paMusBlocks[uiX].uiScriptLen;
	}


	/**
	 * !!!!---- CALL and TRACK LINKING ----!!!!
	 * Scan our list of CALL tags and attempt to resolve their targets in
	 * the binary, conveniently all CALL targets are stored as 32-Bit
	 * signed offsets from the byte after the CALL instruction.
	 *
	 * Once this is done, we'll step through our track list and add the
	 * main tracks to the music binary's directory.
	 */
	for(uiX = 0; uiX < uiTags; uiX++)
	{
		if(mus_getBlock(paMusTags[uiX].pszName, &uiY, &pTarget))
		{
			fprintf(stderr, "%s%s in : %s,%u", szMUS_LostBlock,
				paMusTags[uiX].pszName, paMusTags[uiX].pszSrcFile,
				paMusTags[uiX].uiLine);
			goto cleanup;
		}

		uiZ = (pTarget->uiOffset - paMusTags[uiX].uiOffset) - sizeof(uint32_t);
		pCall = pMusic + uiBlockBase + paMusTags[uiX].uiOffset;
		memcpy(pCall, &uiZ, sizeof(uint32_t));
	}

	for(uiX = 0; uiX < uiTracks; uiX++)
	{
		uiY = uiBlockBase + paMusTracks[uiX].uiMainOffset;

		pDir[CORETONE_MUSPAK_ENTRY_PRIORITY] = paMusTracks[uiX].iMainPriority;
		memcpy(pDir + CORETONE_MUSPAK_ENTRY_OFFSET, &uiY, sizeof(uint32_t));
		pDir += CORETONE_MUSPAK_ENTRY_LEN;
	}


	/**
	 * !!!!---- WRITE BINARY ----!!!!
	 * Do that.
	 */
	uiY = fwrite(pMusic, sizeof(uint8_t), uiLen, pDest);
	if(uiLen != uiY)
	{
		fprintf(stderr, "%s%s\n", szMUS_BadWrite, pszFileName);
		goto cleanup;
	}

	fprintf(stdout, "\nSuccess! %u Bytes written to : %s\n", uiY, pszFileName);
	fprintf(stdout, "Total Tracks %u, Total Blocks : %u\n", uiTracks, uiBlocks);
	iStatus = 0;
	fflush(pDest);

cleanup:
	if(NULL != pMusic) free(pMusic);

	return iStatus;
}

/* void mus_close(void)
 *  Close the muisc block, tag, and track descriptor tables. Call once prior
 * to the exit of the program.
 *----------------------------------------------------------------------------*/
void mus_close(void)
{
	uint32_t uiX;


	if(NULL != pszMusBuf)
	{
		free(pszMusBuf);
		pszMusBuf = NULL;
	}

	if(NULL != paMusBlocks)
	{
		for(uiX = 0; uiX <= uiBlocks; uiX++)
		{
			if(NULL != paMusBlocks[uiX].pszName)
			{
				free(paMusBlocks[uiX].pszName);
				paMusBlocks[uiX].pszName = NULL;
			}

			if(NULL != paMusBlocks[uiX].pacStream)
			{
				free(paMusBlocks[uiX].pacStream);
				paMusBlocks[uiX].pacStream = NULL;
			}
		}

		free(paMusBlocks);
		paMusBlocks = NULL;
	}

	if(NULL != paMusTags)
	{
		/**
		 * While blocks and channels use their last entry (indexed by
		 * the count directly, rather than - 1) for the current work
		 * in progress element, tags are always added in one step
		 * so we don't need to check each of their members.
		 */
		for(uiX = 0; uiX < uiTags; uiX++)
		{
			free(paMusTags[uiX].pszName);
			paMusTags[uiX].pszName = NULL;

			free(paMusTags[uiX].pszSrcFile);
			paMusTags[uiX].pszSrcFile = NULL;
		}

		free(paMusTags);
		paMusTags = NULL;
	}

	if(NULL != paMusTracks)
	{
		free(paMusTracks);
		paMusTracks = NULL;
	}
}




/******************************************************************************
 * !!!!----                  MUSIC DATA MANAGEMENT                   ----!!!!
 ******************************************************************************/
/* int32_t mus_addByte(MusicBlock_t *pBlock, uint8_t ucByte)
 *  Add the given byte to this music block's data stream, returning nonzero
 * if any errors occur.
 *----------------------------------------------------------------------------*/
int32_t mus_addByte(MusicBlock_t *pBlock, uint8_t ucByte)
{
	if(tealloc(&(pBlock->pacStream), &(pBlock->uiBufferLen),
		(pBlock->uiScriptLen + 1), sizeof(uint8_t))) return -1;

	pBlock->pacStream[pBlock->uiScriptLen] = ucByte;
	pBlock->uiScriptLen++;
	uiMusOffset++;
	return 0;
}

/* int32_t mus_prepBlock(MusicBlock_t *pBlock, char *pszName)
 *  Prep the given music block for use, allocating space for its script,
 * allocating and copying in its name, then defaulting all parameters.
 *----------------------------------------------------------------------------*/
int32_t mus_prepBlock(MusicBlock_t *pBlock, char *pszName)
{
	pBlock->pszName = NULL;
	pBlock->pacStream = NULL;
	pBlock->iWasWaiting = 0;
	pBlock->uiScriptLen = 0;
	pBlock->uiBufferLen = 0;
	pBlock->uiOffset = uiMusOffset;

	pBlock->pszName = malloc(sizeof(char) * (strlen(pszName) + 1));
	if(NULL == pBlock->pszName)
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}

	pBlock->pacStream = malloc(sizeof(uint8_t) * MUSIC_PARSER_STREAM_LEN);
	if(NULL == pBlock->pacStream)
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}

	strcpy(pBlock->pszName, pszName);
	pBlock->uiBufferLen = MUSIC_PARSER_STREAM_LEN;
	return 0;
}

/* int32_t mus_getBlock(char *pszName, uint32_t *puiNum, MusicBlock_t **ppBlock)
 *  Attempt to locate and fetch a music block with the name pszName, returning
 * its index and descriptor entry in *puiNum and *ppBlock, respectively.
 *
 *  Returns a nonzero value if the block could not be located.
 *----------------------------------------------------------------------------*/
int32_t mus_getBlock(char *pszName, uint32_t *puiNum, MusicBlock_t **ppBlock)
{
	int32_t iFailed = -1;
	uint32_t uiX;

	if(NULL != paMusBlocks)
	{
		for(uiX = 0; uiX < uiBlocks; uiX++)
		{
			if(0 == strcmp(pszName, paMusBlocks[uiX].pszName))
			{
				if(NULL != puiNum)
					*puiNum = uiX;
				if(NULL != ppBlock)
					*ppBlock = paMusBlocks + uiX;

				iFailed = 0;
				break;
			}
		}
	}

	return iFailed;
}

/* int32_t mus_addCallTag(char *pszTarget, uint32_t uiOffset,
 *                        char *pszSrcFile, uint32_t uiLine)
 *  Add a call reference to pszTarget given the current offset, file, and
 * parsing line, returns nonzero if any errors occur.
 *----------------------------------------------------------------------------*/
int32_t mus_addCallTag(char *pszTarget, uint32_t uiOffset,
                       char *pszSrcFile, uint32_t uiLine)
{
	if(tealloc(&paMusTags, &uiTagSlots, (uiTags + 1), sizeof(MusicTag_t)))
		return -1;

	paMusTags[uiTags].pszName = malloc(sizeof(char) * (strlen(pszTarget) + 1));
	if(NULL == paMusTags[uiTags].pszName)
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}

	paMusTags[uiTags].pszSrcFile = malloc(sizeof(char) * (strlen(pszSrcFile) + 1));
	if(NULL == paMusTags[uiTags].pszSrcFile)
	{
		free(paMusTags[uiTags].pszName);
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}

	strcpy(paMusTags[uiTags].pszName, pszTarget);
	strcpy(paMusTags[uiTags].pszSrcFile, pszSrcFile);
	paMusTags[uiTags].uiOffset = uiMusOffset;
	paMusTags[uiTags].uiLine = uiMusLine;
	uiTags++;
	return 0;
}
