/******************************************************************************
 * wintone.cpp
 * Windows + DirectSound wrappers for CoreTone.
 *-----------------------------------------------------------------------------
 * Version 1.2cz, March 3rd, 2016
 * Copyright (C) 2015 - 2016 Osman Celimli
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 ******************************************************************************/
#include <windows.h>
#include <dsound.h>
#include <stdint.h>
#include <stdio.h>

#include "../types.h"
#include "wintone.hpp"

extern "C" {
	#include "../CoreTone/coretone.h"
};

/******************************************************************************
 * !!!!----          SYSTEM STATE AND DIRECTSOUND RESOURCES          ----!!!!
 ******************************************************************************/
WAVEFORMATEX ToneFormat =
{
	WAVE_FORMAT_PCM,									/* wFormatTag */
	2,													/* nChannels */
	CORETONE_RENDER_RATE,								/* nSamplesPerSec */
	(CORETONE_RENDER_RATE * 2 * sizeof(int16_t)),		/* nAvgBytesPerSec */
	((2 * 16) / 8),										/* nBlockAlign */
	16,													/* wBitsPerSample */
	0													/* cbSize */
};

DSBUFFERDESC ToneBuffer_PriDesc =
{
	sizeof(DSBUFFERDESC),								/* dwSize */
	DSBCAPS_PRIMARYBUFFER,								/* dwFlags */
	0,													/* dwBufferBytes */
	0,													/* dwReserved */
	NULL												/* lpwfxFormat */
};

DSBUFFERDESC ToneBuffer_SecDesc =
{
	sizeof(DSBUFFERDESC),								/* dwSize */
	(DSBCAPS_GLOBALFOCUS | DSBCAPS_CTRLPOSITIONNOTIFY
		| DSBCAPS_GETCURRENTPOSITION2),					/* dwFlags */
	(WINTONE_DBUFFER_LEN * sizeof(int16_t)),			/* dwBufferBytes */
	0,													/* dwReserved */
	&ToneFormat											/* lpwfxFormat */
};

LPDIRECTSOUND lpDirectTone = NULL;
LPDIRECTSOUNDBUFFER lpToneBuffer_Pri = NULL;
LPDIRECTSOUNDBUFFER lpToneBuffer_Sec = NULL;

UINT uToneTimer = NULL;
int32_t iFirstRender = 0;
uint32_t uiLastCursor,uiLastWrite;
uint32_t uiWriteLatency,uiWriteStep;

int16_t aRenderBuffer[WINTONE_RBUFFER_LEN];
uint32_t uiToneTicks = 0;
uint32_t uiRenderIn,uiRenderOut;

HWND hToneWindow = NULL;
int32_t iToneReady = 0;

HANDLE hToneCritical = NULL;
HANDLE hToneBusy = NULL;
int32_t iCriticalSet = 0,iBusySet = 0;
int32_t iTimerSet = 0;

FILE *pLogFile = NULL;
int32_t iLogOutput = 0;
uint32_t uiLogLen = 0;


/******************************************************************************
 * !!!!----            UPDATE TICK AND WAVEFORM RENDERING            ----!!!!
 ******************************************************************************/
/* void CALLBACK wt_renderUpdate(UINT uTimerID, UINT uMsg,
 *                               DWORD dwUser, DWORD dw1, DWORD, dw2)
 *  Render update check which repeatedly fires at WINTONE_UPDATE_LATENCY_MS and
 * will prepare the next set of sample data for the DirectSound buffer if its
 * playback position has moved one or more CoreTone render lengths.
 *----------------------------------------------------------------------------*/
void CALLBACK wt_renderUpdate(UINT uTimerID, UINT uMsg,
							  DWORD dwUser, DWORD dw1, DWORD dw2)
{
	VOID *pvDirectBase_A,*pvDirectBase_B;
	DWORD dwDirectBytes_A,dwDirectBytes_B;
	DWORD dwPlayCursor,dwWriteCursor;

	uint32_t uiDiff,uiUpdates,uiFlush;
	uint32_t uiX,uiY,uiZ;
	HRESULT hRes;


	/**
	 * If we've doubly-entered the rendering callback (maybe we're on
	 * a 486SX-33, who knows), bail out immediately so we don't get
	 * a pile of callbacks all trying to enter this critical section.
	 */
	if(WAIT_OBJECT_0 != WaitForSingleObject(hToneBusy, 0))
	{
		return;
	}

	/**
	 * All our pacing is determined by the current position of the play and
	 * write cursors in DirectSound. The play cursor marks the portion of the
	 * audio buffer currently being farted out the speakers, while the write
	 * cursor indicates the START of the safe region AFTER (as in later, as in
	 * to the right, as in more) the play cursor where we can write.
	 *
	 * For exceptional clarity, things look like this:
	 *  < -------------[PLAY CURSOR xxxxxxxxxxxx WRITE CURSOR]--------------- >
	 *
	 * The area between the brackets with the x's is bad. Don't write there.
	 * Anyway, each time this callback fires we check the amount the play
	 * cursor has moved since our last render. If it's been more than the
	 * length of one or more CoreTone updates in samples, we'll render.
	 *
	 * Data are added to an area after the write cursor, offset by a bit since
	 * both cursors are moving during this update and can fluctuate in their
	 * velocities based upon system activity and driver quality.
	 */
	hRes = lpToneBuffer_Sec->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor);
	if(FAILED(hRes)) goto cleanup;

	uiWriteLatency = (dwPlayCursor <= dwWriteCursor)
		? (dwWriteCursor - dwPlayCursor)
		: ((ToneBuffer_SecDesc.dwBufferBytes - dwPlayCursor) + dwWriteCursor);
	if(iFirstRender)
	{
		/**
		 * This is incredibly dirty, but the write latency is chaotic and
		 * likes to change after playback has started. So we're adding a
		 * small safety border to the write cursor.
		 */
		if(uiWriteLatency > (uiWriteStep * WINTONE_WRITE_BORDER))
		{
			uiLastWrite =
				(dwWriteCursor + (uiWriteLatency * WINTONE_LATENCY_PRESCALE))
				% ToneBuffer_SecDesc.dwBufferBytes;
		}
		else
		{
			uiLastWrite = 
				(dwWriteCursor + (uiWriteStep * WINTONE_WRITE_BORDER))
				% ToneBuffer_SecDesc.dwBufferBytes;
		}

		uiLastWrite = ((uiLastWrite / uiWriteStep) + 1) * uiWriteStep;
		uiLastCursor = (dwPlayCursor / uiWriteStep) * uiWriteStep;
		
		uiRenderIn = 0;
		uiRenderOut = 0;
		iFirstRender = 0;
		uiUpdates = WINTONE_UPDATE_TICKS;
	}
	else
	{
		uiDiff = (uiLastCursor <= dwPlayCursor)
			? (dwPlayCursor - uiLastCursor)
			: ((ToneBuffer_SecDesc.dwBufferBytes - uiLastCursor) + dwPlayCursor);	

		uiUpdates = uiDiff / uiWriteStep;
	}

	/**
	 * If we've gotten this callback close enough to the last update that
	 * there's nothing to do, leave immediately. We could have also missed
	 * our update completely due to some pig hogging the system.
	 *
	 * The latter case is more notable since the play cursor has shot far
	 * enough away from us that we need to either scramble to do updates
	 * or wait for the buffer to wrap around.
	 */
	if((0 == uiUpdates) || (uiUpdates > WINTONE_UPDATE_TICKS))
		goto cleanup;

	/**
	 * If the write cursor plus a safety border has eclipsed our render
	 * offset, the system latency may have increased or the callbacks
	 * have started to fall out of sync (but not severely enough to be
	 * caught by the statement above).
	 *
	 * Our write offset will be bumped back to its minimum safety border
	 * in either of these situations so the user will only get a brief
	 * stutter in the audio instead of repeatedly missed updates as
	 * DirectSound refuses to grant us buffer access.
	 */
	uiY = (dwWriteCursor + (uiWriteStep * WINTONE_PUSH_OFFSET))
		% ToneBuffer_SecDesc.dwBufferBytes;
	if(dwPlayCursor <= uiY)
	{
		if((uiLastWrite <= uiY) && (uiLastWrite >= dwPlayCursor))
		{
			uiLastWrite = 
				(dwWriteCursor + (uiWriteStep * WINTONE_WRITE_BORDER))
				% ToneBuffer_SecDesc.dwBufferBytes;
			uiLastWrite = ((uiLastWrite / uiWriteStep) + 1) * uiWriteStep;
		}
	}
	else
	{
		if((uiLastWrite >= dwPlayCursor) && (uiLastWrite <= uiY))
		{
			uiLastWrite = 
				(dwWriteCursor + (uiWriteStep * WINTONE_WRITE_BORDER))
				% ToneBuffer_SecDesc.dwBufferBytes;
			uiLastWrite = ((uiLastWrite / uiWriteStep) + 1) * uiWriteStep;
		}
	}

	/**
	 * Assuming we've gotten this far, we can grab the region of the
	 * DirectSound secondary buffer that we're going to write to.
	 *
	 * Rendering is staggered, so there is already data in the CoreTone
	 * work buffer, aRenderBuffer, for us to write immediately. This is
	 * another circular buffer the size of WINTONE_RBUFFER_LEN, which
	 * we track using uiRenderOut and uiRenderIn (outgoing and newly
	 * rendered offsets respectively).
	 * uiRenderOut leads uiRenderIn, and if they're equal it means all
	 * slots within the buffer must be refreshed.
	 */
	uiX = uiLastWrite;
	uiY = uiUpdates * uiWriteStep;
	uiLastWrite = (uiLastWrite + uiY) % ToneBuffer_SecDesc.dwBufferBytes;
	uiLastCursor = (uiLastCursor + uiY) % ToneBuffer_SecDesc.dwBufferBytes;

	hRes = lpToneBuffer_Sec->Lock(
		uiX, uiY,
		&pvDirectBase_A, &dwDirectBytes_A,
		&pvDirectBase_B, &dwDirectBytes_B, 0);
	if(FAILED(hRes)) goto cleanup;

	/**
	 * Alright, time to do our rendering update which is
	 * divided into two steps :
	 *
	 * 1) Copy the results from the last rendering update into the area
	 * of DirectSound's secondary buffer we've picked out and locked in
	 * the routine above.
	 *
	 * 2) Do the rendering update itself, the results of which will be
	 * added to the buffer on the next update tick.
	 */
	uiX = (uiRenderOut * uiWriteStep) >> 1;
	if(NULL != pvDirectBase_A)
	{
		for(uiZ = 0; uiZ < dwDirectBytes_A; uiZ += uiWriteStep)
		{
			memcpy(((char*)pvDirectBase_A + uiZ),
				(aRenderBuffer + uiX), uiWriteStep);

			uiRenderOut = (uiRenderOut + 1) % WINTONE_UPDATE_TICKS;
			uiX = (uiX + (uiWriteStep >> 1)) % WINTONE_RBUFFER_LEN;
		}
	}
	if(NULL != pvDirectBase_B)
	{
		for(uiZ = 0; uiZ < dwDirectBytes_B; uiZ += uiWriteStep)
		{
			memcpy(((char*)pvDirectBase_B + uiZ),
				(aRenderBuffer + uiX), uiWriteStep);

			uiRenderOut = (uiRenderOut + 1) % WINTONE_UPDATE_TICKS;
			uiX = (uiX + (uiWriteStep >> 1)) % WINTONE_RBUFFER_LEN;
		}
	}

	hRes = lpToneBuffer_Sec->Unlock(
		pvDirectBase_A, dwDirectBytes_A,
		pvDirectBase_B, dwDirectBytes_B);
	if(FAILED(hRes)) goto cleanup;

	/**
	 * !!!!---- CORETONE RENDER CALL ----!!!!
	 * We've taken care of DirectSound. Now we can render the
	 * waveform data to send out on the next update.
	 *
	 * Again, as our update rate is likely to warble, we'll only
	 * take care of the rendering we need to based upon how much
	 * data was flushed out above. The CoreTone render buffer
	 * should stay saturated at all times.
	 */
	if(uiRenderIn == uiRenderOut)
	{
		uiY = WINTONE_UPDATE_TICKS;
	}
	else
	{
		uiY =  (uiRenderIn < uiRenderOut)
			? (uiRenderOut - uiRenderIn)
			: ((WINTONE_UPDATE_TICKS - uiRenderIn) + uiRenderOut);
	}

	if(WAIT_OBJECT_0 != WaitForSingleObject(hToneCritical, INFINITE))
	{
		goto cleanup;
	}
	uiX = (uiRenderIn * uiWriteStep) >> 1;
	for(uiZ = 0; uiZ < uiY; uiZ++)
	{
		ct_update(aRenderBuffer + uiX);
		if(iLogOutput && (NULL != pLogFile) && (uiLogLen < WINTONE_WAVE_LENGTH_MAX))
		{
			uiFlush = fwrite((aRenderBuffer + uiX), sizeof(uint8_t),
				uiWriteStep, pLogFile);

			uiLogLen += uiFlush;
		}

		uiRenderIn = (uiRenderIn + 1) % WINTONE_UPDATE_TICKS;
		uiX = (uiX + (uiWriteStep >> 1)) % WINTONE_RBUFFER_LEN;
		uiToneTicks++;
	}
	ReleaseMutex(hToneCritical);

cleanup:
	ReleaseMutex(hToneBusy);
}




/******************************************************************************
 * !!!!----                GLOBAL PLAYBACK CONTROL                   ----!!!!
 ******************************************************************************/
void wt_pause(void)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_pause();
			ReleaseMutex(hToneCritical);
		}
	}
}

void wt_resume(void)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_resume();
			ReleaseMutex(hToneCritical);
		}
	}
}

int32_t wt_isPaused(void)
{
	int32_t iAmPaused = -1;

	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			iAmPaused = ct_isPaused();
			ReleaseMutex(hToneCritical);
		}
	}

	return iAmPaused;
}

void wt_stopAll(void)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_stopAll();
			ReleaseMutex(hToneCritical);
		}
	}
}

void wt_setRenderCall(wt_renderCall_t pCall)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_setRenderCall(pCall);
			ReleaseMutex(hToneCritical);
		}
	}
}




/******************************************************************************
 * !!!!----                 MUSIC PLAYBACK CONTROL                   ----!!!!
 ******************************************************************************/
void wt_playMusic(uint8_t *pMusic)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_playMusic(pMusic);
			ReleaseMutex(hToneCritical);
		}
	}
}

void wt_stopMusic(void)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_stopMusic();
			ReleaseMutex(hToneCritical);
		}
	}
}

void wt_attenMusic(int8_t cVol)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_attenMusic(cVol);
			ReleaseMutex(hToneCritical);
		}
	}
}

int32_t wt_checkMusic(void)
{
	int32_t iPlaying = 0;

	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			iPlaying = ct_checkMusic();
			ReleaseMutex(hToneCritical);
		}
	}

	return iPlaying;
}

int32_t wt_getMood(void)
{
	int32_t iMood = 0;

	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			iMood = ct_getMood();
			ReleaseMutex(hToneCritical);
		}
	}

	return iMood;
}




/******************************************************************************
 * !!!!----                  SFX PLAYBACK CONTROL                    ----!!!!
 ******************************************************************************/
void wt_playSFX(uint8_t *pSFX, int8_t cPriority, int8_t cVol_Left, int8_t cVol_Right)
{
	if(iToneReady && (0 != cPriority))
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_playSFX(pSFX, cPriority, cVol_Left, cVol_Right);
			ReleaseMutex(hToneCritical);
		}
	}
}

void wt_stopSFX(int8_t cPriority)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_stopSFX(cPriority);
			ReleaseMutex(hToneCritical);
		}
	}
}

void wt_addSFX(uint8_t *pSFX, int8_t cPriority, int8_t cVol_Left, int8_t cVol_Right)
{
	if(iToneReady && (0 != cPriority))
	{
		/**
		 * We don't need to grab the mutex before submitting the sound effect
		 * here, since they're just going into the intermediary batch list
		 * and won't affect the renderer until the user calls wt_dumpSFX().
		 *
		 * So there's two reasons to use this mechanism - sync with the game
		 * tick and remove the bottleneck of waiting for the render mutex
		 * on each sound effect submission.
		 */
		ct_addSFX(pSFX, cPriority, cVol_Left, cVol_Right);
	}
}

void wt_dumpSFX(void)
{
	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			ct_dumpSFX();
			ReleaseMutex(hToneCritical);
		}
	}
}




/******************************************************************************
 * !!!!----                      OUTPUT LOGGING                      ----!!!!
 ******************************************************************************/
/* int32_t wt_logOutput(char *pszFile)
 *  Attempt to begin logging the softsynth output to the given file in the
 * RIFF/WAVE format. Will return a nonzero value if the log could not start.
 *----------------------------------------------------------------------------*/
int32_t wt_logOutput(char *pszFile)
{
	int32_t iStatus = -1;

	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			/**
			 * If we're not already logging, attempt to open the target file
			 * for binary writing and seek past the header. We'll backtrack
			 * and write the RIFF/WAVE info once the log has been closed.
			 */
			if((0 == iLogOutput) && (NULL == pLogFile))
			{
				if(NULL != (pLogFile = fopen(pszFile, "wb")))
				{
					fseek(pLogFile, WINTONE_WAVE_DATA_START, SEEK_SET);
					uiLogLen = WINTONE_WAVE_DATA_START;
					iLogOutput = -1;

					iStatus = 0;
				}
			}

			ReleaseMutex(hToneCritical);
		}
	}
	return iStatus;
}

/* void wt_closeLog(void)
 *  Stop the currently active output log, stamp the RIFF/WAV header on the
 * target file, and close it.
 *----------------------------------------------------------------------------*/
void wt_closeLog(void)
{
	uint32_t uiChunkSize,uiDataSize;
	uint32_t uiX;
	uint16_t usY;

	if(iToneReady
		&& (WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE)))
	{
		if(iLogOutput && (NULL != pLogFile))
		{
			fflush(pLogFile);
			rewind(pLogFile);

			/**
			 * RIFF/WAV Format :
			 * [RIFF Descriptor][Format Sub-Chunk][Data Sub-Chunk] . . .
			 *
			 * RIFF Descriptor :
			 *	00 - 03		ChunkID "RIFF"
			 *	04 - 07		ChunkSize
			 *	08 - 11		Format "WAVE"
			 *
			 * Format Sub-Chunk :
			 *	12 - 15		SubChunk1ID "fmt "
			 *	16 - 19		SubChunk1Size
			 *	20 - 21		AudioFormat
			 *	22 - 23		NumChannels
			 *	24 - 27		SampleRate
			 *	28 - 31		ByteRate
			 *	32 - 33		BlockAlign
			 *	34 - 35		BitsPerSample
			 *
			 * Data Sub-Chunk :
			 *	36 - 39		SubChunk2ID "data"
			 *	40 - 43		SubChunk2Size
			 *	44 +		Data
			 */
			uiChunkSize = uiLogLen - 8;
			uiDataSize = uiLogLen - WINTONE_WAVE_DATA_START;

			fwrite(WINTONE_WAVE_CHUNK_ID, sizeof(uint8_t), 4, pLogFile);
			fwrite(&uiChunkSize, sizeof(uint8_t), 4, pLogFile);
			fwrite(WINTONE_WAVE_CHUNK_FMT, sizeof(uint8_t), 4, pLogFile);

			fwrite(WINTONE_WAVE_SUBCHUNK_F_ID, sizeof(uint8_t), 4, pLogFile);
			uiX = 16;
			fwrite(&uiX, sizeof(uint8_t), 4, pLogFile);
			usY = 1;
			fwrite(&usY, sizeof(uint8_t), 2, pLogFile);
			usY = 2;
			fwrite(&usY, sizeof(uint8_t), 2, pLogFile);
			uiX = CORETONE_RENDER_RATE;
			fwrite(&uiX, sizeof(uint8_t), 4, pLogFile);
			uiX = CORETONE_RENDER_RATE * 2 * sizeof(int16_t);
			fwrite(&uiX, sizeof(uint8_t), 4, pLogFile);
			usY = ((2 * 16) / 8);
			fwrite(&usY, sizeof(uint8_t), 2, pLogFile);
			usY = 16;
			fwrite(&usY, sizeof(uint8_t), 2, pLogFile);

			fwrite(WINTONE_WAVE_SUBCHUNK_D_ID, sizeof(uint8_t), 4, pLogFile);
			fwrite(&uiDataSize, sizeof(uint8_t), 4, pLogFile);

			fflush(pLogFile);
			fclose(pLogFile);

			iLogOutput = 0;
			pLogFile = NULL;
		}

		ReleaseMutex(hToneCritical);
	}
}



/******************************************************************************
 * !!!!----                       MUTEX ACCESS                       ----!!!!
 ******************************************************************************/
/* int32_t wt_getMutex(void)
 *  Aquire the WinTone access mutex, upon which the caller will be guaranteed
 * that neither WinTone nor CoreTone will perform an update until the mutex is
 * released through wt_giveMutex().
 *  This can be used to ensure a user callback routine will not be called
 * while changing its operating parameters. Returns nonzero on failure.
 *----------------------------------------------------------------------------*/
int32_t wt_getMutex(void)
{
	int32_t iFailed = -1;

	if(iToneReady)
	{
		if(WAIT_OBJECT_0 == WaitForSingleObject(hToneCritical, INFINITE))
		{
			iFailed = 0;
		}
	}

	return iFailed;
}

/* int32_t wt_giveMutex(void)
 *  Release the WinTone access mutex, returns nonzero on failure.
 *----------------------------------------------------------------------------*/
int32_t wt_giveMutex(void)
{
	int32_t iFailed = -1;

	if(iToneReady)
	{
		ReleaseMutex(hToneCritical);
		iFailed = 0;
	}

	return iFailed;
}




/******************************************************************************
 * !!!!----               INITIALIZATION & DIAGNOSTICS               ----!!!!
 ******************************************************************************/
/* HRESULT wt_init(HWND hWindow, uint8_t *pSamplePak, uint8_t *pInstrPak)
 *  Initialize both the CoreTone library and DirectSound hookups using the
 * supplied parent window, sample package, and instrument package.
 *  Will return a generic E_FAIL HRESULT if there was an issue initializing
 * CoreTone or the specific Windows + DirectSound error if the problem
 * originated from an interaction with either of those.
 *----------------------------------------------------------------------------*/
HRESULT wt_init(HWND hParent, uint8_t *pSamplePak, uint8_t *pInstrPak)
{
	VOID *pvDirectBase_A,*pvDirectBase_B;
	DWORD dwDirectBytes_A,dwDirectBytes_B;

	int32_t iStatus;
	HRESULT hRes = E_FAIL;


	iToneReady = 0;
	uiLastCursor = 0;

	iTimerSet = 0;
	iBusySet = 0;
	iCriticalSet = 0;
	uiToneTicks = 0;

	pLogFile = NULL;
	iLogOutput = 0;
	uiLogLen = 0;
	if((NULL != hParent) && (NULL != pSamplePak) && (NULL != pInstrPak))
	{
		iStatus = ct_init(pSamplePak, pInstrPak);
		if(0 == iStatus)
		{
			hRes = DirectSoundCreate(NULL, &lpDirectTone, NULL);
			if(FAILED(hRes) || (NULL == lpDirectTone)) goto cleanup;

			hRes = lpDirectTone->SetCooperativeLevel(hParent, DSSCL_PRIORITY);
			if(FAILED(hRes)) goto cleanup;

			/**
			 * DirectSound operates upon sound buffers.
			 * - For each application there is a single PRIMARY buffer
			 * which will be sent to an audio device and farted out
			 * of some speakers.
			 * - For each sound you want to play there is a SECONDARY
			 * buffer which will then be mixed into the PRIMARY buffer.
			 *
			 * We'll perform our waveform rendering in the SECONDARY buffer
			 * which will then be transferred to the PRIMARY buffer.
			 */
			hRes = lpDirectTone->CreateSoundBuffer(
				&ToneBuffer_PriDesc, &lpToneBuffer_Pri, NULL);
			if(FAILED(hRes) || (NULL == lpToneBuffer_Pri)) goto cleanup;

			hRes = lpToneBuffer_Pri->SetFormat(&ToneFormat);
			if(FAILED(hRes)) goto cleanup;

			hRes = lpDirectTone->CreateSoundBuffer(
				&ToneBuffer_SecDesc, &lpToneBuffer_Sec, NULL);
			if(FAILED(hRes) || (NULL == lpToneBuffer_Sec)) goto cleanup;

			/**
			 * Our DirectSound buffers are ready, so let's clear out both
			 * our render and DirectSound buffers, set up our callback,
			 * and kick off circular playback.
			 */
			memset(aRenderBuffer, 0, (WINTONE_RBUFFER_LEN * sizeof(int16_t)));

			hRes = lpToneBuffer_Sec->Lock(0,
				(WINTONE_DBUFFER_LEN * sizeof(int16_t)),
				&pvDirectBase_A, &dwDirectBytes_A,
				&pvDirectBase_B, &dwDirectBytes_B, DSBLOCK_ENTIREBUFFER);
			if(FAILED(hRes)) goto cleanup;

			if(NULL != pvDirectBase_A)
				memset(pvDirectBase_A, 0, dwDirectBytes_A);
			if(NULL != pvDirectBase_B)
				memset(pvDirectBase_B, 0, dwDirectBytes_B);

			hRes = lpToneBuffer_Sec->Unlock(pvDirectBase_A, dwDirectBytes_A,
				pvDirectBase_B, dwDirectBytes_B);
			if(FAILED(hRes)) goto cleanup;

			/**
			 * We'll be monitoring the playback cursor in regular intervals
			 * using a multimedia timer in order to decide when to perform
			 * our render updates. See wt_renderUpdate() for details.
			 *
			 * Originally I was using two position notifiers, but these are
			 * much less accurate than desired and don't work at all on
			 * certain hardware + drivers which I happen to own and use.
			 * This has generally improved on other (less old) machines, but
			 * we need this library to be compatible.
			 *
			 * Access to CoreTone + WinTone's resources will be protected
			 * by the mutexes hToneBusy and hToneCritical, as the render
			 * callback will run in its own thread.
			 */
			lpToneBuffer_Sec->SetCurrentPosition(0);
			iFirstRender = -1;
			uiWriteStep = CORETONE_BUFFER_LEN * sizeof(int16_t);

			hToneBusy = CreateMutex(NULL, FALSE, NULL);
			if(NULL == hToneBusy)
			{
				hRes = E_FAIL;
				goto cleanup;
			}
			iBusySet = -1;

			hToneCritical = CreateMutex(NULL, FALSE, NULL);
			if(NULL == hToneCritical)
			{
				hRes = E_FAIL;
				goto cleanup;
			}
			iCriticalSet = -1;

			/**
			 * I'm starting playback before the multimedia timer since the
			 * callback routine would very much like to know about buffer
			 * latency, which requires activity.
			 */
			hRes = lpToneBuffer_Sec->Play(0, 0, DSBPLAY_LOOPING);
			if(FAILED(hRes)) goto cleanup;

			uToneTimer = timeSetEvent(WINTONE_UPDATE_LATENCY_MS,
				WINTONE_UPDATE_RESOLUTION_MS, wt_renderUpdate,
				NULL,
				(TIME_PERIODIC | TIME_CALLBACK_FUNCTION));
			if(NULL == uToneTimer)
			{
				hRes = HRESULT_FROM_WIN32(GetLastError());
				goto cleanup;
			}
			else
			{
				iTimerSet = -1;
			}

			iToneReady = -1;
		}
	}

cleanup:
	if(FAILED(hRes))
	{
		wt_close();
	}

	return hRes;
}

/* void wt_close(void)
 *  Close down DirectSound before we leave.
 *----------------------------------------------------------------------------*/
void wt_close(void)
{
	iToneReady = 0;

	/**
	 * Cease playback on our secondary buffer, release our timer event,
	 * close any pending log files, and make sure the render routine has
	 * terminated before doing anything else.
	 */
	if(NULL != lpToneBuffer_Sec)
	{
		lpToneBuffer_Sec->Stop();
	}

	if(NULL != uToneTimer)
	{
		timeKillEvent(uToneTimer);
		uToneTimer = NULL;
	}

	wt_closeLog();
	if(iBusySet)
	{
		WaitForSingleObject(hToneBusy, INFINITE);
	}
	if(iCriticalSet)
	{
		WaitForSingleObject(hToneCritical, INFINITE);
	}

	/**
	 * Now we can release all of the DirectSound buffers
	 * and interfaces.
	 */
	if(NULL != lpToneBuffer_Sec)
	{
		lpToneBuffer_Sec->Release();
		lpToneBuffer_Sec = NULL;
	}

	if(NULL != lpToneBuffer_Pri)
	{
		lpToneBuffer_Pri->Release();
		lpToneBuffer_Pri = NULL;
	}

	if(NULL != lpDirectTone)
	{
		lpDirectTone->Release();
		lpDirectTone = NULL;
	}

	if(iCriticalSet)
	{
		ReleaseMutex(hToneCritical);
		CloseHandle(hToneCritical);

		hToneCritical = NULL;
		iCriticalSet = 0;
	}
	if(iBusySet)
	{
		ReleaseMutex(hToneBusy);
		CloseHandle(hToneBusy);

		hToneBusy = NULL;
		iBusySet = 0;
	}
}

/* void wt_getState(void **ppaChannels, void **ppaPatches, void **ppaTracks,
 *                  int32_t *piToneReady, uint32_t *puiTicks,
 *                  int16_t **ppaRenderWave, uint32_t *puiRenderLen)
 *  As with most functions in WinTone this is just a wrapper, in this case
 * around ct_getState. However, it features three additions - whether or not
 * the library is ready (0 != *piToneReady), the number of update ticks
 * performed since initialization (*puiTicks), and the base address of the
 * final waveform render buffer (**ppaRenderWave).
 *----------------------------------------------------------------------------*/
void wt_getState(void **ppaChannels, void **ppaPatches, void **ppaTracks,
                 int32_t *piToneReady, uint32_t *puiTicks,
                 int16_t **ppaRenderWave, uint32_t *puiRenderLen)
{
	ct_getState(ppaChannels, ppaPatches, ppaTracks);

	if(NULL != piToneReady)
		*piToneReady = iToneReady;
	if(NULL != puiTicks)
		*puiTicks = uiToneTicks;
	if(NULL != ppaRenderWave)
		*ppaRenderWave = aRenderBuffer;
	if(NULL != puiRenderLen)
		*puiRenderLen = WINTONE_RBUFFER_LEN;
}

/* int32_t wt_getInfo(uint32_t *puiChannels, uint32_t *puiRenderFreq,
 *                    uint32_t *puiDecodeRate,
 *                    uint32_t *puiSamples, uint32_t *puiSampleLen)
 *  Get extended information about the WinTone + CoreTone version including
 * their channel count, render frequency, decode rate, maximum sample
 * package size, and sample length. Also verifies the CoreTone and WinTone
 * builds are matched in capabilities, returning a nonzero value if there are
 * any discrepencies which could affect playback.
 *
 *  This should be used to distinguish among variations in WinTone + CoreTone
 * DLLs and ensure the loaded versions are right ones for your software.
 *----------------------------------------------------------------------------*/
int32_t wt_getInfo(uint32_t *puiChannels, uint32_t *puiRenderFreq,
                   uint32_t *puiDecodeRate,
                   uint32_t *puiSamples, uint32_t *puiSampleLen)
{
	uint32_t uiC,uiR,uiD,uiS,uiL;
	int32_t iMisMatch = 0;

	ct_getInfo(&uiC, &uiR, &uiD, &uiS, &uiL);
	iMisMatch |= (CORETONE_CHANNELS == uiC) ? 0 : -1;
	iMisMatch |= (CORETONE_RENDER_RATE == uiR) ? 0 : -1;
	iMisMatch |= (CORETONE_DECODE_RATE == uiD) ? 0 : -1;

	if(NULL != puiChannels)
		*puiChannels = uiC;
	if(NULL != puiRenderFreq)
		*puiRenderFreq = uiR;

	if(NULL != puiDecodeRate)
		*puiDecodeRate = uiD;

	if(NULL != puiSamples)
		*puiSamples = uiS;
	if(NULL != puiSampleLen)
		*puiSampleLen = uiL;

	return iMisMatch;
}
