/******************************************************************************
 * bupboop.h
 * The "final" user-facing portions of BupBoop and the CoreTone SoftSynth.
 *
 * All functions pertaining to music and sound playback are wrapped here with
 * bb_x prefixes. These should be called by your application instead of one of
 * the platform specific functions (such as WinTone's wt_x) in order to
 * increase portability.
 *
 * Anything related to library setup or teardown is expected to remain platform
 * specific due to the vast differences among the supported operating systems
 * and embedded devices. These definitions can be excluded when not required
 * by defining BUPBOOP_EXCLUDE_OS_SPECIFIC prior to including this file.
 *-----------------------------------------------------------------------------
 * Version 1.2cz, March 3rd, 2016
 * Copyright (C) 2015 - 2016 Osman Celimli
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 ******************************************************************************/
#ifndef BUPBOOP
#define BUPBOOP

/******************************************************************************
 * Preface
 ******************************************************************************/
/* Which files to include and where audio calls will be remapped is chosen
 * based upon the following :
 *
 * - If _WIN32 is defined, WinTone will be used.
 * - If none of the above are true, CoreTone is assumed to be called directly.
 */
#if defined(_WIN32)
	#include "WinTone/wintone.hpp"
#else
	#include "CoreTone/coretone.h"
#endif


/* ---- TYPING ----
 * bb_renderCall_t
 *  Post-render callback.
 */
#if defined(_WIN32)
	#define bb_renderCall_t wt_renderCall_t
#else
	#define bb_renderCall_t ct_renderCall_t
#endif


/******************************************************************************
 * Function Defines
 ******************************************************************************/
/* ---- GLOBAL PLAYBACK CONTROL ----
 * void bb_pause(void)
 *  Pause (and silence) the decoding of all active music and sound effects.
 *
 * void bb_resume(void)
 *  Resume audio playback from a paused state.
 *
 * int32_t bb_isPaused(void)
 *  Indicates if audio is currently paused (nonzero) or unpaused (zero).
 *
 * void bb_stopAll(void)
 *  Halt the decoding and playback off all sound effects and music.
 */
#if defined(_WIN32)
	#define bb_pause wt_pause
	#define bb_resume wt_resume
	#define bb_isPaused wt_isPaused
	#define bb_stopAll wt_stopAll
#else
	#define bb_pause ct_pause
	#define bb_resume ct_resume
	#define bb_isPaused ct_isPaused
	#define bb_stopAll ct_stopAll
#endif


/* ---- MUSIC PERTINENT FUNCTIONS ----
 * void bb_playMusic(uint8_t *pMusic)
 *  Reqest the playback of the music track whose base address is at pMusic
 * on the next update tick. Will "prestop" the current playing track (if any).
 *
 * void bb_stopMusic(void)
 *  Request any currently playing music to cease on the next update tick.
 *
 * void bb_attenMusic(int8_t cVol)
 *  Request a change in volume for the currently playing music (if any) on
 * the next update tick, zero is silent and 127 is loudest.
 *
 * int32_t bb_checkMusic(void)
 *  Check to see whether the music is playing (nonzero return) or not (zero),
 * useful to determine if a single-shot track has finished or if a stop
 * request has completed.
 *
 * int32_t bb_getMood(void)
 *  Get the mood flag of the currently playing music track. If no music is
 * playing, the mood flag will be zero (neutral).
 */
#if defined(_WIN32)
	#define bb_playMusic(pMusic) wt_playMusic(pMusic)
	#define bb_stopMusic wt_stopMusic
	#define bb_attenMusic(cVol) wt_attenMusic(cVol)
	#define bb_checkMusic wt_checkMusic
	#define bb_getMood wt_getMood
#else
	#define bb_playMusic(pMusic) ct_playMusic(pMusic)
	#define bb_stopMusic ct_stopMusic
	#define bb_attenMusic(cVol) ct_attenMusic(cVol)
	#define bb_checkMusic ct_checkMusic
	#define bb_getMood ct_getMood
#endif


/* ---- SOUND EFFECT PERTINENT FUNCTIONS ----
 * void bb_playSFX(uint8_t *pSFX, int8_t cPriority,
 *                 int8_t cVol_Left, int8_t cVol_Right)
 *  Request the playback of the sound effect pSFX on the next update tick with
 * priority ucPriority and panning cVol_Left and cVol_Right.
 *
 * void bb_stopSFX(uint8_t ucPriority)
 *  Request all currently decoding sound effects with the priority ucPriority
 * to cease playback on the next update tick.
 *
 * void bb_playSFX(uint8_t *pSFX, int8_t cPriority,
 *                 int8_t cVol_Left, int8_t cVol_Right)
 *  Request the playback of the sound effect pSFX on the next batch dump with
 * priority ucPriority and panning cVol_Left and cVol_Right. This can be used
 * to synchronize the start of multiple sound effects.
 *
 * void bb_dumpSFX(void)
 *  Play all sound effects in the current batch set using ct_playSFX(), this
 * can be used to synchronize the start of multiple sound effects.
 */
#if defined(_WIN32)
	#define bb_playSFX(pSFX, cPriority, cVol_Left, cVol_Right) wt_playSFX(pSFX, cPriority, cVol_Left, cVol_Right)
	#define bb_stopSFX(cPriority) wt_stopSFX(cPriority)
	#define bb_addSFX(pSFX, cPriority, cVol_Left, cVol_Right) wt_addSFX(pSFX, cPriority, cVol_Left, cVol_Right)
	#define bb_dumpSFX wt_dumpSFX
#else
	#define bb_playSFX(pSFX, cPriority, cVol_Left, cVol_Right) ct_playSFX(pSFX, cPriority, cVol_Left, cVol_Right)
	#define bb_stopSFX(cPriority) ct_stopSFX(cPriority)
	#define bb_addSFX(pSFX, cPriority, cVol_Left, cVol_Right) ct_addSFX(pSFX, cPriority, cVol_Left, cVol_Right)
	#define bb_dumpSFX ct_dumpSFX
#endif


/* ---- USER HOOKS and DRIVER CONTROL ----
 * void bb_setRenderCall(pCall)
 *  Set the given function pCall as a render callback which will be called every
 * CoreTone update and given access to the waveform buffer.
 *
 * int32_t bb_getMutex(void)
 *  Acquire the render mutex, which will block CoreTone from performing an update
 * tick until it is released.
 *
 * int32_t bb_giveMutex(void)
 *  Release the render mutex.
 */
#if defined(_WIN32)
	#define bb_setRenderCall(pCall) wt_setRenderCall(pCall)
	#define bb_getMutex wt_getMutex
	#define bb_giveMutex wt_giveMutex
#else
	#define bb_setRenderCall(pCall) ct_setRenderCall(pCall)
	#define bb_getMutex ct_getMutex
	#define bb_giveMutex ct_giveMutex
#endif

#endif
